#include "pwatch_com.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>

#include "pwatch_proc.h"


static unsigned int	pwatch_version = 0;

/*
 * Check if fdinfo folder is available 
 * Returns 0 if yes, -1 if no 
 */
int pwatch_fdinfo_avail(void)
{
	DIR *d;
	char dir[128];

	snprintf(dir, sizeof(dir), "/proc/%d/fdinfo", getpid());
	d = opendir(dir);
	if (!d)
		return -1;

	closedir(d);

	return 0;
}

/*
 * Check if /proc/<pid>/io folder is available 
 * Returns 0 if yes, -1 if no 
 */
int pwatch_io_avail(void)
{
	int fd;
	char file[128];

	snprintf(file, sizeof(file), "/proc/%d/io", getpid());
	fd = open(file, O_RDONLY);
	if (fd == -1)
		return -1;

	close(fd);

	return 0;
}

/*
 * Load info about a fd
 */
static int pwatch_proc_fd_info(const pid_t pid, const int fd, struct pwatch_fd *ifd)
{
	char f[128];
	char r[1024], *p;
	ssize_t bytes;
	struct stat st;

	ifd->fd = fd;

	/* fd */
	snprintf(f, sizeof(f), "/proc/%ld/fd/%d",
		(long) pid, fd);
	bytes = readlink(f, ifd->file, sizeof(ifd->file) - 1);
	ifd->file[bytes] = '\0';

	if (stat(f, &st) == 0)
		ifd->size = st.st_size;
	else
		ifd->size = 0;

	/* fdinfo */
	snprintf(f, sizeof(f), "/proc/%ld/fdinfo/%d",
		(long) pid, fd);
	if (pwatch_get_content(f, r, sizeof(r)) < 0)
		return -1;

	ifd->offset = 0;
	p = strstr(r, "pos:");
	if (p)
		ifd->offset = strtoll(p + 5, NULL, 10);

	p = strstr(r, "flags:");
	if (p)
		ifd->flags = strtol(p + 7, NULL, 8);

	return 0;
}

/*
 * Outputs pwatch_fd structure as text
 */
void pwatch_fd_str(const struct pwatch_fd *ifd, char *buf, const size_t buf_len)
{
	if (!ifd) {
		strcpy(buf, "");
		return;
	}

	snprintf(buf, buf_len, "?"
		);
}

/*
 * Removes a fd list from a proc struct
 */
static int pwatch_proc_unload_fds(struct pwatch_proc *proc)
{
	struct pwatch_fd *f, *next;

	f = proc->fd_head;
	while (f) {
		next = f->next;
		free(f);
		f = next;
	}

	proc->fd_head = NULL;

	return 0;
}

/*
 * Adds a fd to tree
 */
static int pwatch_proc_add_fd(struct pwatch_proc *proc,
	const struct pwatch_fd *ifd)
{
	struct pwatch_fd *p, *q, *prev;

	p = proc->fd_head;
	q = NULL;
	prev = NULL;
	while (p) {
		if (p->fd == ifd->fd) {
			q = p;
			break;
		}

		prev = p;
		p = p->next;
	}

	if (!q) {
		q = (struct pwatch_fd *) malloc(sizeof(struct pwatch_fd));
		if (!q)
			return -1;

		memset(q, 0, sizeof(struct pwatch_fd));

		q->next = NULL;

		if (prev) {
			prev->next = q;
		} else {
			proc->fd_head = q;
		}

		q->fd = ifd->fd;
	}

	snprintf(q->file, sizeof(q->file), "%s", ifd->file);
	q->offset = ifd->offset;
	q->size = ifd->size;
	q->flags = ifd->flags;

	return 0;
}

/*
 * Loads a tree of fds
 */
int pwatch_proc_load_fds(struct pwatch_proc *proc)
{
	char dir[256];
	DIR *d;
	struct dirent *e;
	struct pwatch_fd ifd;
	int fd;

	snprintf(dir, sizeof(dir), "/proc/%d/fd", proc->pid);
	d = opendir(dir);
	if (!d)
		return -1;

	while ((e = readdir(d))) {
		if (e->d_name[0] == '.')
			continue;

		fd = strtol(e->d_name, NULL, 10);

		if (pwatch_proc_fd_info(proc->pid, fd, &ifd) < 0) {
			closedir(d);
			return -1;
		}

		/* filter some stuff */
		if (strncmp(ifd.file, "/proc", 5) == 0)
			continue;
		if (strncmp(ifd.file, "/dev/null", 9) == 0)
			continue;
		if (strncmp(ifd.file, "pipe:", 5) == 0)
			continue;
		if (strncmp(ifd.file, "socket:", 7) == 0)
			continue;
		if (strncmp(ifd.file, "anon_inode:", 11) == 0)
			continue;

		if (pwatch_proc_add_fd(proc, &ifd) < 0) {
			closedir(d);
			return -1;
		}
	}
	closedir(d);

	return 0;
}

/*
 * Add a process to the list
 */
static int pwatch_proc_add(struct pwatch_proc **head,
	const struct pwatch_proc *proc)
{
	struct pwatch_proc *p, *q, *prev;
	struct timeval now;
	long int diff_in_ms;

	if (!head)
		return -1;

	p = *head;
	prev = NULL;
	q = NULL;
	while (p) {
		if (p->pid == proc->pid) {
			q = p;
			break;
		}

		prev = p;
		p = p->next;
	}

	if (!q) {
		q = (struct pwatch_proc *) malloc(sizeof(struct pwatch_proc));
		if (!q)
			return -1;

		memset(q, 0, sizeof(struct pwatch_proc));

		q->next = NULL;

		if (prev) {
			prev->next = q;
		} else {
			*head = q;
		}

		q->pid = proc->pid;
		q->speed_read = 0;
		q->speed_write = 0;
	}

	snprintf(q->name, sizeof(q->name), "%s", proc->name);

	/* Compute speed */
	gettimeofday(&now, NULL);
	if (q->last_set.tv_sec > 0) {
		diff_in_ms = ((now.tv_sec - q->last_set.tv_sec) * 1000000
			+ (now.tv_usec - q->last_set.tv_usec)) / 1000;
		q->speed_read = (proc->ts.read_bytes - q->ts.read_bytes)
			/ diff_in_ms * 1000;
		q->speed_write = (proc->ts.write_bytes - q->ts.write_bytes)
			/ diff_in_ms * 1000;
	}
	q->last_set = now;

	/* Set new values */
	memcpy(&q->ts, &proc->ts, sizeof(struct taskstats));

	q->version = pwatch_version;

	return 0;
}

/*
 * Load processes informations
 */
int pwatch_proc_load(struct pwatch_proc **head)
{
	DIR *d;
	struct dirent *e;
	char file[256];
	char info[1024];
	struct pwatch_proc proc;
	char *s, *se;
	int procs = 0;
	pid_t pid;

	d = opendir("/proc");
	if (!d)
		return -1;

	pwatch_version = 1 - pwatch_version;

	while ((e = readdir(d))) {
		memset(&proc, 0, sizeof(proc));

		/* We skip non-process objects */
		proc.pid = strtoul(e->d_name, NULL, 10);
		if (proc.pid == 0)
			continue;

		/* Find out the process name */
		snprintf(file, sizeof(file), "/proc/%s/cmdline", e->d_name);
		if (pwatch_get_content_text(file, proc.name, sizeof(proc.name)) < 0)
			continue;

		if (strlen(proc.name) == 0) {
			snprintf(file, sizeof(file), "/proc/%s/status", e->d_name);
			if (pwatch_get_content(file, info, sizeof(info)) < 0)
				continue;
			s = strstr(info, "Name:");
			if (s) {
				se = strchr(s, '\n');
				if (se) {
					*se = '\0';
					snprintf(proc.name, sizeof(proc.name), "[%s]",
						s + 6);
				}
			}
		}

		/* Get info about I/O using netlink */
		pid = strtol(e->d_name, NULL, 10);
		if (pwatch_nl_get_taskstats(pid, &proc.ts) < 0) {
			/* signal error */
			proc.ts.read_bytes = 1;
			proc.ts.write_bytes = 1;
			continue;
		}

		if (pwatch_proc_add(head, &proc) < 0)
			continue;

		procs++;
	}
	closedir(d);

	return procs;
}

/*
 * Dump statistics to a file
 */
void pwatch_proc_dump2file(const struct pwatch_proc *head)
{
	const struct pwatch_proc *p;
	int out;
	char file[128];
	char buf[1024];
	struct timeval now;

	gettimeofday(&now, NULL);

	snprintf(file, sizeof(file), "io_%ld.dump",
		now.tv_sec);

	out = open(file, O_WRONLY | O_CREAT | O_EXCL | O_TRUNC, 0600);
	if (out == -1)
		return;

	snprintf(buf, sizeof(buf),
		"ProcessID\tBytes read\tSpeed read\t"
		"Bytes write\tSpeed write\tProcess name\n");
	write(out, buf, strlen(buf));

	p = head;
	while (p) {
		snprintf(buf, sizeof(buf), "%9u"
			"\t%llu"
			"\t%llu"
			"\t%llu"
			"\t%llu"
			"\t%s\n",
			p->pid, p->ts.read_bytes, p->speed_read,
			p->ts.write_bytes, p->speed_write,
			p->name);

		write(out, buf, strlen(buf));

		p = p->next;
	}
	close(out);
}

/*
 * Clean terminated processes
 */
void pwatch_proc_clean(struct pwatch_proc **head)
{
	struct pwatch_proc *p, *next, *prev;

	if (!head)
		return;

	p = *head;
	prev = NULL;
	while (p) {
		next = p->next;

		if (p->version != pwatch_version) {
			if (prev) {
				prev->next = next;
			} else {
				*head = next;
			}

			pwatch_proc_unload_fds(p);
			free(p);
		} else {
			prev = p;
		}

		p = next;
	}
}

/*
 * Sort the tree
 * TODO: Better sorting algo, please!
 */
void pwatch_proc_sort(struct pwatch_proc **head, const unsigned int sort_by,
	const unsigned int inverse)
{
	struct pwatch_proc *p, *next, *prev;
	int stop = 0;
	unsigned int switches = 0;
	unsigned long long int tmp, v1, v2;

	if (!head)
		return;

	while (stop == 0) {
		stop = 1;
		p = *head;
		prev = NULL;
		while (p) {
			next = p->next;
			if (!next)
				break;

			switch (sort_by) {
				case SORT_BY_BYTES_READ:
					v1 = p->ts.read_bytes;
					v2 = next->ts.read_bytes;
					break;
				case SORT_BY_BYTES_WRITE:
					v1 = p->ts.write_bytes;
					v2 = next->ts.write_bytes;
					break;
				case SORT_BY_SPEED_READ:
					v1 = p->speed_read;
					v2 = next->speed_read;
					break;
				case SORT_BY_SPEED_WRITE:
					v1 = p->speed_write;
					v2 = next->speed_write;
					break;
			}

			if (inverse == 1) {
				tmp = v1;
				v1 = v2;
				v2 = tmp;
			}

			if (v1 < v2) {
				if (prev)
					prev->next = next;
				else
					*head = next;
				p->next = next->next;
				next->next = p;
				stop = 0;
				switches++;
			}

			prev = p;
			p = p->next;
		}
	}
}

/*
 * Returns a proc by searching for pid
 */
struct pwatch_proc *pwatch_proc_get(struct pwatch_proc *head,
	const pid_t pid)
{
	while (head) {
		if (head->pid == pid)
			return head;
		head = head->next;
	}

	return NULL;
}
